/**
 * A PBKDF2 routine for use in window and web workers.
 * Generates a function that tries Web Crypto and falls back to asmCrypto.
 * Requires asmCrypto.
 * @param object [global] window in the browser or self in a web worker.
 */
var KeeperPBKDF2 = function (global) {
    
    global = global || self;
    
    var crypto, subtle;
  
    // Safari Version 10.1.1 (12603.2.4) throws NotSupportedError accessing webkitSubtle
    try {
      crypto = global.crypto || global.msCrypto;
      subtle = crypto.subtle || crypto.webkitSubtle;
    } catch (err) {
      console.log(new Error(err));
    }
  
    /**
     * Is a native WebCrypto-based PBKDF2 operation is supported?
     */
    var nativeSupport = (function () {
      var password =  new Uint8Array([65 /* random byte */ ]);
      return new Promise(function (resolve, reject) {
        // Safari throws an error without returning a promise.
        // IE/Edge resolves the promise but throws an error 'PBKDF2'.
        try {
          return subtle.importKey('raw', password, 'PBKDF2', false, [ 'deriveBits' ])
            .then(function () {
              resolve(true);
            })
            .catch(function (err) {
              resolve(false);
            });
        } catch (err) {
          resolve(false);
        }
      });
    }());
  
    var nativePBKDF2 = function (password, saltBytes, iterations, byteLength) {
  
      var encoder = new TextEncoder('utf-8');
      
      var passwordBuf = encoder.encode(password);
  
      // Old versions of Safari don't return a rejected promise, but throw an unhandled exception.
      // We handle that case with this try/catch construct.
      try {
        return subtle.importKey('raw', passwordBuf, 'PBKDF2', false, [ 'deriveBits' ])
          .then(function (key) {
            var params = {
              name: 'PBKDF2',
              salt: saltBytes,
              iterations: iterations,
              hash: {
                name: 'SHA-256'
              }
            };
  
            return subtle.deriveBits(params, key, (byteLength || 32) * 8);
          });
      } catch (err) {
        return Promise.reject(new Error(err));
      }
    }
  
    var asmcryptoPBKDF2 = function (password, saltBytes, iterations, byteLength) {
      return new Promise(function (resolve, reject) {
        var hash = asmCrypto.PBKDF2_HMAC_SHA256.base64(unescape(encodeURIComponent(password)), saltBytes, iterations, byteLength || 32).replace(/\+/g, '-').replace(/\//g, '_').replace(/=/g, '');
        resolve(hash);
      });
    }
  
    return function (password, saltBase64UrlSafe, iterations) {
      
      return nativeSupport.then(function (supported) {
  
        var saltBase64 = (saltBase64UrlSafe || '').replace(/-/g, '+').replace(/_/g, '/'); // Unnecessary to pad with equal signs when converting back to non-urlsafe base64.
        var saltBytes = asmCrypto.base64_to_bytes(saltBase64);
        var byteLength = 32;
  
        if (supported) {
          return nativePBKDF2(password, saltBytes, iterations, byteLength)
            .then(function (key) {
              return asmCrypto.bytes_to_base64(new Uint8Array(key)).replace(/\+/g, '-').replace(/\//g, '_').replace(/=/g, '');
            })
            .catch(function (err) {
              console.log(new Error(err));
              return asmcryptoPBKDF2(password, saltBytes, iterations, byteLength);
            });
  
        }
        
        return asmcryptoPBKDF2(password, saltBytes, iterations, byteLength);
      });
    };
  
  };